/*
 * Modified and rewriten by Samule Behan <behan@frida.fri.utc.sk> 
 * from original pam_filter by Andrew Morgan <morgan@transmeta.com> and
 * Richard Stevens' UNIX Network Programming book.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <syslog.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <termios.h>
#include <errno.h>
#include <signal.h>

#define PAM_SM_AUTH
#define PAM_SM_ACCOUNT
#define PAM_SM_SESSION
#define PAM_SM_PASSWORD

#include <security/pam_modules.h>

/* log errors */
#include <stdarg.h>

/*log program actions*/
static void _pam_log(int err, const char *format, ...)
{
    va_list args;

    va_start(args, format);
    openlog("pam_watch", LOG_CONS | LOG_PID, LOG_USER);	/*LOG_AUTH */
    vsyslog(err, format, args);
    va_end(args);
    closelog();
}

#define TERMINAL_LEN 12
/*find and open master terminal*/
static int master(char *terminal)
{
    const char ptys[] = "pqrs", *pty = ptys;
    const char hexs[] = "0123456789abcdef", *hex;
    struct stat tstat;
    int fd;

    strcpy(terminal, "/dev/pty??");
    while (*pty)
    {				/* step through four types */
	terminal[8] = *pty++;
	terminal[9] = '0';
	if (stat(terminal, &tstat) < 0)
	{
	    _pam_log(LOG_WARNING, "unknown pseudo terminal: %s", terminal);
	    break;
	}
	for (hex = hexs; *hex;)
	{			/* step through 16 of these */
	    terminal[9] = *hex++;
	    if ((fd = open(terminal, O_RDWR)) >= 0)
		return fd;
	}
    }
    /* no terminal found, sorry */
    return -1;
}

/*process module arguments*/
#define WATCH_DEBUG	01
#define WATCH_PATH	02
#define WATCH_NOSPY	04
#define WATCH_NOWRITE	010
#define NEW_TERM	100
#define NON_TERM	200
#define NO_WARN		400

static int process_args(pam_handle_t * pamh, int argc, char **argv,
			char **socket_path, char **nospy)
{
    int ctrl = 0;

    while (argc-- > 0)
    {
	if (strcmp("debug", *argv) == 0)	/*debug mode */
	    ctrl |= WATCH_DEBUG;
	else if (strcmp("nowrite", *argv) == 0)	/*disable write access */
	    ctrl |= WATCH_NOWRITE;
	else if (strcmp("new_term", *argv) == 0)	/*want new term - unused */
	    ctrl |= NEW_TERM;
	else if (strcmp("non_term", *argv) == 0)	/*want no term  - unused */
	    ctrl |= NON_TERM;
	else if (strcmp("no_warn", *argv) == 0)	/*no warnings  - unused */
	    ctrl |= NO_WARN;
	else if (strstr(*argv, "nospy=") != NULL)	/*do not spy this */
	{
	    ctrl |= WATCH_NOSPY;
	    *nospy = *argv + 6;
	}
	else if (strstr(*argv, "path=") != NULL)	/*socket path */
	{
	    ctrl |= WATCH_PATH;
	    *socket_path = *argv + 5;
	}
	else
	    _pam_log(LOG_ERR, "unrecognized option: %s (ignored)", *argv);
	++argv;
    }

    if (!(ctrl & WATCH_PATH))
    {
	_pam_log(LOG_ERR, "socket path argument required");
	return -1;
    }

    return ctrl;
}

static int set_watch(pam_handle_t * pamh, int flags, int ctrl,
		     const char *socket_path)
{
    int pid;
    int fd[2];
    struct termios old_mode;	/* initial terminal mode settings */
    char terminal[TERMINAL_LEN];

    /*check socket path and path existence */
    if ((socket_path == NULL) || (*socket_path != '/'))
    {
	if (ctrl & WATCH_DEBUG)
	    _pam_log(LOG_ALERT, "invalid socket path");
	return PAM_ABORT;
    }
    else
    {				/*check before any of important operation is done */
	struct stat dstat;
	if ((stat(socket_path, &dstat) == -1) || (!S_ISDIR(dstat.st_mode)))
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_ALERT, "socket path '%s' does not exists",
			 socket_path);
	    return PAM_ABORT;
	}
    }

    /*create master->slave terminal interlink */
    if ((fd[0] = master(terminal)) == -1)
    {
	if (ctrl & WATCH_DEBUG)
	    _pam_log(LOG_ALERT, "can not set master terminal (%s)",
		     strerror(errno));
	return PAM_ABORT;
    }
    terminal[5] = 't';
    if ((fd[1] = open(terminal, O_RDWR)) == -1)
    {
	if (ctrl & WATCH_DEBUG)
	    _pam_log(LOG_ALERT, "can not set slave terminal (%s)",
		     strerror(errno));
	return PAM_ABORT;
    }

    /*get terminal mode and set it raw */
    if (tcgetattr(STDIN_FILENO, &old_mode) == 0)
    {
	struct termios t_mode = old_mode;

	cfmakeraw(&t_mode);	/*make raw */
	if (tcsetattr(STDIN_FILENO, TCSAFLUSH, &t_mode) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_ALERT, "can not put terminal to raw mode");
	    return PAM_ABORT;
	}
    }
    else
    {
	if (ctrl & WATCH_DEBUG)
	    _pam_log(LOG_ALERT, "can not copy terminal mode");
	return PAM_ABORT;
    }

/* ----------------------------------------------------------------------------
*	-------------------- Watching parent ------------------
*  ---------------------------------------------------------------------------- */
    if ((pid = fork()) && (pid != -1))
    {
	fd_set readers;
	char buffer[BUFSIZ];
	short int open_watch = 0, user_nowrite = 0;
	int fifo_in = -1, fifo_out = -1;
	int APPIN_FILENO = 3, APPOUT_FILENO = 4, APPERR_FILENO = 5;
	char *fifo_name = NULL, *user_name = NULL, *service_name = NULL;

	/*helper functions */
	void catch_pipe()
	    { if (fifo_out != -1)
		close(fifo_out);
	    fifo_out = -1;
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING,
			 "watcher disconnected from output fifo");
	}
	void exit_cleanup()
	    { if (tcsetattr(STDIN_FILENO, TCSANOW, &old_mode) == -1)
	    {
		if (ctrl & WATCH_DEBUG)
		    _pam_log(LOG_WARNING,
			     "can not set back terminal mode on exit");
	    }
	    unlink(fifo_name);
	    if (!(ctrl & WATCH_NOWRITE))
	    {
		strcat(fifo_name, "-i");
		unlink(fifo_name);
	    }
	    kill(pid, 0);	/*preventive kill, not realy needed */
	}
	void open_output_pipe()
	    { if (fifo_out == -1)
		if ((fifo_out = open(fifo_name, O_WRONLY | O_NONBLOCK)) !=
		    -1)
		{
		    if (ctrl & WATCH_DEBUG)
			_pam_log(LOG_WARNING,
				 "watcher connected to output fifo");
		    open_watch = 0;
		}
	    signal(13, *catch_pipe);
	}
	void open_input_pipe()
	    { if (fifo_in != -1)
		close(fifo_in);
	    else
		return;
	    fifo_in = -1;
	    if (
		(fifo_in =
		 open(strcat(fifo_name, "-i"), O_RDONLY | O_NONBLOCK)) == -1)
	    {
		errno = 0;	/*we dont care about input, output is important */
		if (ctrl & WATCH_DEBUG)
		    _pam_log(LOG_WARNING, "can not re-open input fifo '%s'",
			     fifo_name);}
	    fifo_name[strlen(fifo_name) - 2] = '\0';	/*remove '-i' */
	}
	void catch_usr1()
	    { if (fifo_out == -1)
		open_watch = 1;
	    signal(10, *catch_usr1);
	}
	void catch_usr2()
	{ user_nowrite = !user_nowrite;
	}
	/*register on exit fucntion */ if (atexit(*exit_cleanup) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING, "can not register at exit function");
	    exit_cleanup();
	    exit(0);
	}

/*BUGFIX: login program automatically starts timeout function, and this
          kills our parent login shell so disable alarm
*/
       alarm((unsigned int)0);

	/*setup master side interlink */
	close(fd[1]);
	if ((APPIN_FILENO = dup(fd[0])) == -1
	    || (APPOUT_FILENO = dup(fd[0])) == -1
	    || (APPERR_FILENO = dup(fd[0])) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING,
			 "can not assign controling STDIN/OUT/ERR");
	    exit(0);
	}
	close(fd[0]);
	if ((fcntl(APPIN_FILENO, F_SETFD, 0)) == -1
	    || (fcntl(APPOUT_FILENO, F_SETFD, 0)) == -1
	    || (fcntl(APPERR_FILENO, F_SETFD, 0)) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING,
			 "unable to retain controling I/O modes");
	    exit(0);
	}

	/*initialize spy fifos */
	if (chdir(socket_path) != 0)
	{
	    _pam_log(LOG_ERR, "socket path '%s' not accessible",
		     socket_path);
	    exit(errno);
	}

	/*create and cd fifo dir */
	pam_get_item(pamh, PAM_USER, (void *) &user_name);
	if (mkdir(user_name, 0700) == -1 && (errno != EEXIST))
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_CRIT,
			 "can not create dir '%s/%s' for user '%s'",
			 socket_path, user_name, user_name);
	    exit(errno);
	}
	if (chdir(user_name) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_CRIT, "can not cd for %s", user_name);
	    exit(errno);
	}

	/*create output fifo */
	pam_get_item(pamh, PAM_SERVICE, (void *) &service_name);
	if ((fifo_name = (char *) malloc(sizeof(char) * (strlen(service_name)
							 + strlen(terminal) +
							 10))) == NULL)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_CRIT, "can not malloc at fifo_name");
	    exit(errno);
	}
	sprintf(fifo_name, "%s-%d.%s", service_name, getpid(), terminal + 5);
	if (mkfifo(fifo_name, 0600) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING, "unable to create output fifo");
	    exit(errno);
	}

	/*create input fifo */
	if (!(ctrl & WATCH_NOWRITE))
	{
	    strcat(fifo_name, "-i");
	    if (mkfifo(fifo_name, 0600) == -1)
	    {
		if (ctrl & WATCH_DEBUG)
		    _pam_log(LOG_WARNING, "unable to create input fifo");
		exit(errno);
	    }
	    if ((fifo_in = open(fifo_name, O_RDONLY | O_NONBLOCK)) == -1)
	    {
		errno = 0;	/*we dont care about input, output is important */
		if (ctrl & WATCH_DEBUG)
		    _pam_log(LOG_WARNING, "can not open input fifo '%s'",
			     fifo_name);
	    }
	    fifo_name[strlen(fifo_name) - 2] = '\0';	/*remove '-i' */
	}

	/*dealloc everythink (look for pam) */
	free(user_name);
	free(service_name);
	(void) pam_set_item(pamh, PAM_AUTHTOK, NULL);
	(void) pam_set_item(pamh, PAM_OLDAUTHTOK, NULL);

	/*install watch signal - SIGUSR1, SIGUSR2 */
	signal(10, *catch_usr1);
	signal(12, *catch_usr2);
	
	/*start I/O copying & watching */
	FD_ZERO(&readers);
	while (1)
	{
	    /*open output fifo if reader avaible (only on SIGUSR1) */
	    if ((fifo_out == -1) && (open_watch))
		open_output_pipe();

	    /*dancing round the console I/O */
	    FD_SET(APPOUT_FILENO, &readers);
	    FD_SET(APPERR_FILENO, &readers);
	    FD_SET(STDIN_FILENO, &readers);
	    if (!(ctrl & WATCH_NOWRITE) && (fifo_in != -1))
		FD_SET(fifo_in, &readers);

/*ADDME?: Here can be screen fifo be opened, because SIGUSR1 usualy breaks
      the select. I dont want to do it, because it does not look very good 
      (for me) but if anybody want add it to FD_ZERO() command.
    On the other hand, client application is not usualy so fast, it can
      open screen fifo immediatelly after sending SIGUSR1 signal...
  NOTE: I have made it 'open on signal only', because else I would need 
      in each loop to check if somebody is hearing on screen fifo and 
      that's bad.
*/
	    if (select(APPERR_FILENO + 2, &readers, NULL, NULL, NULL) < 0)
	    {
		if (errno != EINTR)
		{
		    if (ctrl & WATCH_DEBUG)
			_pam_log(LOG_WARNING, "select failed (%s)",
				 strerror(errno));
		    break;
		}
		else		/*if signall is received it blames the select, so clean it */
		    FD_ZERO(&readers);
	    }

	    /*user output pipe */
	    if (FD_ISSET(APPOUT_FILENO, &readers))
	    {
		int got = read(APPOUT_FILENO, buffer, BUFSIZ);
		if (got != -1)
		{
		    write(STDOUT_FILENO, buffer, got);
		    if (fifo_out != -1)
			write(fifo_out, buffer, got);
		}
		else
		    break;
	    }
/*      else if(FD_ISSET(APPERR_FILENO, &readers))
      {
        int got	= read(APPERR_FILENO, buffer, BUFSIZ);
	if(got != -1)
	{ write(STDERR_FILENO, buffer, got); }
	else break;
      }   */

/*ADDME: here can be add watching of non echoed user input 
        (ie. too stole passwords - bad example ;)
*/

	    /*user input pipe */
	    if (FD_ISSET(STDIN_FILENO, &readers))
	    {
		int got = read(STDIN_FILENO, buffer, BUFSIZ);
		if (got != -1)
		{
		    if (!user_nowrite)
			write(APPIN_FILENO, buffer, got);
		}
		else
		    break;
	    }

	    /*watcher input pipe */
	    if (!(ctrl & WATCH_NOWRITE) && (fifo_in != -1)
		&& FD_ISSET(fifo_in, &readers))
	    {
		int got = read(fifo_in, buffer, BUFSIZ);
		if (got != -1)
		{
		    if (got == 0)
			open_input_pipe();
		    write(APPIN_FILENO, buffer, got);
		}
		else
		    break;
	    }
	}
	exit(0);
    }
/* ----------------------------------------------------------------------------
*	-------------------- Watched child ------------------
*  ---------------------------------------------------------------------------- */
    else if (pid != -1)
    {
	struct winsize win;

	/*get current winsize */
	if (ioctl(STDOUT_FILENO, TIOCGWINSZ, &win) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING, "can not get current window size");
	    /*we don need to abort this is not very important to terminal */
	}

	/*setup slave side interlink */
	close(fd[0]);
	if (dup2(fd[1], STDIN_FILENO) == -1
	    || dup2(fd[1], STDOUT_FILENO) == -1
	    || dup2(fd[1], STDERR_FILENO) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING, "can not assign STDIN/OUT/ERR");
	    return PAM_ABORT;
	}
	close(fd[1]);
	if (fcntl(STDIN_FILENO, F_SETFD, 0) == -1
	    || fcntl(STDOUT_FILENO, F_SETFD, 0) == -1
	    || fcntl(STDERR_FILENO, F_SETFD, 0) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING, "unable to retain I/O modes");
	    return PAM_ABORT;
	}

	/*became new session leader */
	if (setsid() == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING,
			 "child can not became new session leader");
	    return PAM_ABORT;
	}

	/*get our control terminal, back*/
	if(ioctl(STDIN_FILENO, TIOCSCTTY, STDIN_FILENO))
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING,
			 "child can not set controling terminal, ignoring");
	}

	/*put user terminal to normal mode */
	if (tcsetattr(STDIN_FILENO, TCSANOW, &old_mode) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING, "can not set back terminal mode");
	    return PAM_ABORT;
	}

	/*re-set winsize */
	if (ioctl(STDOUT_FILENO, TIOCSWINSZ, &win) == -1)
	{
	    if (ctrl & WATCH_DEBUG)
		_pam_log(LOG_WARNING, "can not get current window size");
	    /*we don need to abbort this is not very important to terminal */
	}

	/*go back to user */
	return PAM_SUCCESS;
    }

    /*this happens only if fork failed */
    if (ctrl & WATCH_DEBUG)
	_pam_log(LOG_ERR, "fork failed (%s)", strerror(errno));

    return PAM_ABORT;
}

static int set_the_terminal(pam_handle_t * pamh)
{
    const char *tty;

    if (pam_get_item(pamh, PAM_TTY, (const void **) &tty) != PAM_SUCCESS
	|| tty == NULL)
    {
	tty = ttyname(STDIN_FILENO);
	if (tty == NULL)
	{
	    _pam_log(LOG_ERR, "couldn't get the tty name");
	    return PAM_ABORT;
	}
	if (pam_set_item(pamh, PAM_TTY, tty) != PAM_SUCCESS)
	{
	    _pam_log(LOG_ERR, "couldn't set tty name");
	    return PAM_ABORT;
	}
    }
    return PAM_SUCCESS;
}

static int add_watch(pam_handle_t * pamh, int flags, int argc,
		     const char **argv, const char *name)
{
    int ctrl, retval = PAM_SUCCESS;
    char *socket_path, *nospy;

#warning IGNORE next Warning
    ctrl = process_args(pamh, argc, (char **) argv, &socket_path, &nospy);
    if (ctrl == -1)
	return PAM_AUTHINFO_UNAVAIL;

    /*set terminal name */
    if (!(ctrl & NON_TERM) && !(ctrl & NEW_TERM))
    {
	retval = set_the_terminal(pamh);
	if (retval != PAM_SUCCESS)
	    _pam_log(LOG_ERR, "tried and failed to set PAM_TTY");
    }

    /*some users should not be watched */
    if (ctrl & WATCH_NOSPY)
    {
	char *user, *uuser;

	pam_get_item(pamh, PAM_USER, (void *) &user);
	uuser = (char *) malloc(strlen(user) + 2);
	if (uuser != NULL)
	{
	    strcpy(uuser, user);
	    strcat(uuser, ",");
	    if (strstr(nospy, user) != NULL)
	    {
		free(uuser);
		return PAM_SUCCESS;
	    }
	    uuser[strlen(uuser) - 1] = '\0';
	    sprintf(uuser, ",%s", user);
	    if (strstr(nospy, uuser) != NULL)
	    {
		free(uuser);
		return PAM_SUCCESS;
	    }
	}
	if (ctrl & WATCH_DEBUG)
	    _pam_log(LOG_NOTICE, "user %s will be watched", user);
    }

    if (retval == PAM_SUCCESS)
	retval = set_watch(pamh, flags, ctrl, socket_path);

    /*set terminal name */
    if (retval == PAM_SUCCESS && !(ctrl & NON_TERM) && (ctrl & NEW_TERM))
    {
	retval = set_the_terminal(pamh);
	if (retval != PAM_SUCCESS)
	    _pam_log(LOG_ERR,
		     "tried and failed to set new terminal as PAM_TTY");
    }

    if (ctrl & WATCH_DEBUG)
	_pam_log(LOG_DEBUG, "[%s]", pam_strerror(pamh, retval));

    return retval;
}

/* ----------------- public functions ---------------- */

/*
 * here are the advertised access points ...
 */

/* ------------------ authentication ----------------- */

PAM_EXTERN int pam_sm_authenticate(pam_handle_t * pamh, int flags,
				   int argc, const char **argv)
{
    return add_watch(pamh, flags, argc, argv, "authenticate");
}

PAM_EXTERN int pam_sm_setcred(pam_handle_t * pamh, int flags, int argc,
			      const char **argv)
{
    return add_watch(pamh, flags, argc, argv, "setcred");
}

/* --------------- account management ---------------- */

PAM_EXTERN int pam_sm_acct_mgmt(pam_handle_t * pamh, int flags, int argc,
				const char **argv)
{
    return add_watch(pamh, flags, argc, argv, "setcred");
}

/* --------------- session management ---------------- */

PAM_EXTERN int pam_sm_open_session(pam_handle_t * pamh, int flags,
				   int argc, const char **argv)
{
    return add_watch(pamh, flags, argc, argv, "open_session");
}

/*NOTE: There is nothing to be done on session close 
  so dont delay us */

PAM_EXTERN int pam_sm_close_session(pam_handle_t * pamh, int flags,
				    int argc, const char **argv)
{
    return PAM_SUCCESS;
}


/* --------- updating authentication tokens --------- */


PAM_EXTERN int pam_sm_chauthtok(pam_handle_t * pamh, int flags, int argc,
				const char **argv)
{
    return add_watch(pamh, flags, argc, argv, "chauthtok");
}

#ifdef PAM_STATIC

/* ------------ stuff for static modules ------------ */

struct pam_module _pam_watch_modstruct = {
    "pam_watch",
    pam_sm_authenticate,
    pam_sm_setcred,
    pam_sm_acct_mgmt,
    pam_sm_open_session,
    pam_sm_close_session,
    pam_sm_chauthtok,
};

#endif
